@@ -65,6 +65,8 @@
 
 #define PROGNAME        "trurl"
 
+#define REPLACE_NULL_BYTE '.' /* for query:key extractions */
+
 struct var {
   const char *name;
   CURLUPart part;
@@ -99,6 +101,7 @@ static const struct var variables[] = {
 #define ERROR_TRIM   8 /* a --trim problem */
 #define ERROR_BADURL 9 /* if --verify is set and the URL cannot parse */
 #define ERROR_GET   10 /* bad --get syntax */
+#define ERROR_ITER  11 /* bad --iterate syntax */
 
 #ifndef SUPPORTS_URL_STRERROR
 /* provide a fake local mockup */
@@ -141,6 +144,7 @@ static void help(void)
           "  -f, --url-file [file/-]      - read URLs from file or stdin\n"
           "  -g, --get [{component}s]     - output component(s)\n"
           "  -h, --help                   - this help\n"
+          "  --iterate [component]=[list] - create multiple URL outputs\n"
           "  --json                       - output URL as JSON\n"
           "  --query-separator [letter]   - if something else than '&'\n"
           "  --redirect [URL]             - redirect to this\n"
@@ -168,12 +172,21 @@ static void show_version(void)
   exit(0);
 }
 
+struct iterinfo {
+  CURLU *uh;
+  const char *part;
+  size_t plen;
+  char *ptr;
+  unsigned int varmask; /* sets 1 << [component] */
+};
+
 struct option {
   struct curl_slist *url_list;
   struct curl_slist *append_path;
   struct curl_slist *append_query;
   struct curl_slist *set_list;
   struct curl_slist *trim_list;
+  struct curl_slist *iter_list;
   const char *redirect;
   const char *qsep;
   const char *format;
@@ -278,6 +291,15 @@ static void setadd(struct option *o,
     o->set_list = n;
 }
 
+static void iteradd(struct option *o,
+                    const char *iter) /* [component]=[data] */
+{
+  struct curl_slist *n;
+  n = curl_slist_append(o->iter_list, iter);
+  if(n)
+    o->iter_list = n;
+}
+
 static void trimadd(struct option *o,
                     const char *trim) /* [component]=[data] */
 {
@@ -331,6 +353,10 @@ static int getarg(struct option *op,
     setadd(op, arg);
     *usedarg = 1;
   }
+  else if(checkoptarg("--iterate", flag, arg)) {
+    iteradd(op, arg);
+    *usedarg = 1;
+  }
   else if(checkoptarg("--redirect", flag, arg)) {
     if(op->redirect)
       errorf(ERROR_FLAG, "only one --redirect is supported");
@@ -379,38 +405,57 @@ static int getarg(struct option *op,
   return 0;
 }
 
-static void showqkey(const char *key, size_t klen, bool urldecode)
+static void showqkey(const char *key, size_t klen, bool urldecode,
+                     bool showall)
 {
   int i;
+  bool shown = false;
+  char **qp = urldecode ? qpairsdec : qpairs;
+
   for(i = 0; i< nqpairs; i++) {
-    if(urldecode) {
-      if(!strncmp(key, qpairsdec[i], klen) &&
-         (qpairsdec[i][klen] == '=')) {
-        fputs(&qpairsdec[i][klen + 1], stdout);
+    if(!strncmp(key, qp[i], klen) &&
+       (qp[i][klen] == '=')) {
+      if(shown)
+        fputc(' ', stdout);
+      fputs(&qp[i][klen + 1], stdout);
+      if(!showall)
         break;
-      }
-    }
-    else {
-      if(!strncmp(key, qpairs[i], klen) &&
-         (qpairs[i][klen] == '=')) {
-        fputs(&qpairs[i][klen + 1], stdout);
-        break;
-      }
+      shown = true;
     }
   }
 }
 
+/* component to variable pointer */
+static const struct var *comp2var(const char *name, size_t vlen)
+{
+  int i;
+  for(i = 0; variables[i].name; i++)
+    if((strlen(variables[i].name) == vlen) &&
+       !strncmp(name, variables[i].name, vlen))
+      return &variables[i];
+  return NULL;
+}
+
 static void get(struct option *op, CURLU *uh)
 {
   FILE *stream = stdout;
   const char *ptr = op->format;
   bool done = false;
+  char startbyte = 0;
+  char endbyte = 0;
 
   while(ptr && *ptr && !done) {
-    if('{' == *ptr) {
-      if('{' == ptr[1]) {
+    if(!startbyte && (('{' == *ptr) || ('[' == *ptr))) {
+      startbyte = *ptr;
+      if('{' == *ptr)
+        endbyte = '}';
+      else
+        endbyte = ']';
+    }
+    if(startbyte == *ptr) {
+      if(startbyte == ptr[1]) {
         /* an escaped {-letter */
-        fputc('{', stream);
+        fputc(startbyte, stream);
         ptr += 2;
       }
       else {
@@ -418,12 +463,12 @@ static void get(struct option *op, CURLU *uh)
         char *end;
         char *cl;
         size_t vlen;
-        int i;
         bool urldecode = true;
-        end = strchr(ptr, '}');
+        end = strchr(ptr, endbyte);
         ptr++; /* pass the { */
         if(!end) {
           /* syntax error */
+          fputc(startbyte, stream);
           continue;
         }
         /* {path} {:path} */
@@ -432,47 +477,47 @@ static void get(struct option *op, CURLU *uh)
           ptr++;
         }
         vlen = end - ptr;
-        /* check for a colon within here */
+        /* check for the last colon within here */
         cl = memchr(ptr, ':', vlen);
         if(cl) {
           /* deduct the colon part */
-          if(!strncmp(ptr, "query:", 6))
-            showqkey(&ptr[6], end - cl - 1, urldecode);
+          if(!strncmp(ptr, "query-all:", 10))
+            showqkey(&ptr[10], end - cl - 1, urldecode, true);
+          else if(!strncmp(ptr, "query:", 6))
+            showqkey(&ptr[6], end - cl - 1, urldecode, false);
           else
             errorf(ERROR_GET, "Bad --get syntax: %s", ptr);
         }
         else {
-          for(i = 0; variables[i].name; i++) {
-            if((strlen(variables[i].name) == vlen) &&
-               !strncmp(ptr, variables[i].name, vlen)) {
-              char *nurl;
-              CURLUcode rc;
-              rc = curl_url_get(uh, variables[i].part, &nurl,
-                                CURLU_DEFAULT_PORT|
-                                CURLU_NO_DEFAULT_PORT|
-                                (urldecode?CURLU_URLDECODE:0));
-              switch(rc) {
-              case CURLUE_OK:
-                fprintf(stream, "%s", nurl);
-                curl_free(nurl);
-              case CURLUE_NO_SCHEME:
-              case CURLUE_NO_USER:
-              case CURLUE_NO_PASSWORD:
-              case CURLUE_NO_OPTIONS:
-              case CURLUE_NO_HOST:
-              case CURLUE_NO_PORT:
-              case CURLUE_NO_QUERY:
-              case CURLUE_NO_FRAGMENT:
+          const struct var *v = comp2var(ptr, vlen);
+          if(v) {
+            char *nurl;
+            CURLUcode rc;
+            rc = curl_url_get(uh, v->part, &nurl,
+                              CURLU_DEFAULT_PORT|
+                              CURLU_NO_DEFAULT_PORT|
+                              (urldecode?CURLU_URLDECODE:0));
+            switch(rc) {
+            case CURLUE_OK:
+              fprintf(stream, "%s", nurl);
+              curl_free(nurl);
+            case CURLUE_NO_SCHEME:
+            case CURLUE_NO_USER:
+            case CURLUE_NO_PASSWORD:
+            case CURLUE_NO_OPTIONS:
+            case CURLUE_NO_HOST:
+            case CURLUE_NO_PORT:
+            case CURLUE_NO_QUERY:
+            case CURLUE_NO_FRAGMENT:
 #ifdef SUPPORTS_ZONEID
-              case CURLUE_NO_ZONEID:
+            case CURLUE_NO_ZONEID:
 #endif
-                /* silently ignore */
-                break;
-              default:
-                fprintf(stderr, PROGNAME ": %s (%s)\n", curl_url_strerror(rc),
-                        variables[i].name);
-                break;
-              }
+              /* silently ignore */
+              break;
+            default:
+              fprintf(stderr, PROGNAME ": %s (%s)\n", curl_url_strerror(rc),
+                      v->name);
+              break;
             }
           }
         }
@@ -490,6 +535,15 @@ static void get(struct option *op, CURLU *uh)
       case 't':
         fputc('\t', stream);
         break;
+      case '\\':
+        fputc('\\', stream);
+        break;
+      case '{':
+        fputc('{', stream);
+        break;
+      case '[':
+        fputc('[', stream);
+        break;
       default:
         /* unknown, just output this */
         fputc(*ptr, stream);
@@ -506,47 +560,52 @@ static void get(struct option *op, CURLU *uh)
   fputc('\n', stream);
 }
 
+static const struct var *setone(CURLU *uh, struct option *o,
+                                const char *setline)
+{
+  char *ptr = strchr(setline, '=');
+  const struct var *v = NULL;
+  (void)o;
+  if(ptr && (ptr > setline)) {
+    size_t vlen = ptr - setline;
+    bool urlencode = true;
+    bool found = false;
+    if(ptr[-1] == ':') {
+      urlencode = false;
+      vlen--;
+    }
+    v = comp2var(setline, vlen);
+    if(v) {
+      CURLUcode rc;
+      rc = curl_url_set(uh, v->part, ptr[1] ? &ptr[1] : NULL,
+                        CURLU_NON_SUPPORT_SCHEME|
+                        (urlencode ? CURLU_URLENCODE : 0) );
+      if(rc)
+        warnf("Error setting %s: %s", v->name, curl_url_strerror(rc));
+      found = true;
+    }
+    if(!found)
+      errorf(ERROR_SET, "unknown component: %.*s", (int)vlen, setline);
+  }
+  else
+    errorf(ERROR_SET, "invalid --set syntax: %s", setline);
+  return v;
+}
+
 static void set(CURLU *uh,
                 struct option *o)
 {
   struct curl_slist *node;
-  bool varset[NUM_COMPONENTS];
-  memset(varset, 0, sizeof(varset));
+  unsigned int mask = 0;
   for(node =  o->set_list; node; node = node->next) {
+    const struct var *v;
     char *set = node->data;
-    int i;
-    char *ptr = strchr(set, '=');
-    if(ptr && (ptr > set)) {
-      size_t vlen = ptr-set;
-      bool urlencode = true;
-      bool found = false;
-      if(ptr[-1] == ':') {
-        urlencode = false;
-        vlen--;
-      }
-      for(i = 0; variables[i].name; i++) {
-        if((strlen(variables[i].name) == vlen) &&
-           !strncmp(set, variables[i].name, vlen)) {
-          CURLUcode rc;
-          if(varset[i])
-            errorf(ERROR_SET, "A component can only be set once per URL (%s)",
-                   variables[i].name);
-          rc = curl_url_set(uh, variables[i].part, ptr[1] ? &ptr[1] : NULL,
-                            CURLU_NON_SUPPORT_SCHEME|
-                            (urlencode ? CURLU_URLENCODE : 0) );
-          if(rc)
-            warnf("Error setting %s: %s", variables[i].name,
-                  curl_url_strerror(rc));
-          found = true;
-          varset[i] = true;
-          break;
-        }
-      }
-      if(!found)
-        errorf(ERROR_SET, "Set unknown component: %s", set);
+    v = setone(uh, o, set);
+    if(v) {
+      if(mask & (1 << v->part))
+        errorf(ERROR_SET, "duplicate --set for component %s", v->name);
+      mask |= (1 << v->part);
     }
-    else
-      errorf(ERROR_SET, "invalid --set syntax: %s", set);
   }
 }
 
@@ -616,6 +675,7 @@ static void json(struct option *o, CURLU *uh)
         fputs(",\n", stdout);
       printf("    \"%s\": ", variables[i].name);
       jsonString(stdout, nurl, 0, false);
+      curl_free(nurl);
     }
   }
   if(nqpairs) {
@@ -709,10 +769,19 @@ static char *memdupdec(char *source, size_t len)
 
   left = curl_easy_unescape(NULL, source, sep ? (size_t)(sep - source) : len,
                             &leftlen);
-  if(sep)
+  if(sep) {
+    char *p;
+    int plen;
     right = curl_easy_unescape(NULL, sep + 1, len - (sep - source) - 1,
                                &rightlen);
 
+    /* convert null bytes to periods */
+    for(plen = rightlen, p = right; plen; plen--, p++) {
+      if(!*p)
+        *p = REPLACE_NULL_BYTE;
+    }
+  }
+
   str = curl_maprintf("%.*s%s%.*s", leftlen, left,
                       right ? "=":"",
                       rightlen, right?right:"");
@@ -819,10 +888,13 @@ static void sortquery(struct option *o)
 }
 
 static void singleurl(struct option *o,
-                      const char *url) /* might be NULL */
+                      const char *url, /* might be NULL */
+                      struct iterinfo *iinfo,
+                      struct curl_slist *iter)
 {
-    struct curl_slist *p;
-    CURLU *uh = curl_url();
+  CURLU *uh = iinfo->uh;
+  if(!uh) {
+    uh = curl_url();
     if(!uh)
       errorf(ERROR_MEM, "out of memory");
     if(url) {
@@ -843,9 +915,73 @@ static void singleurl(struct option *o,
                        CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME);
       }
     }
+  }
+  do {
+    char iterbuf[1024];
+    struct curl_slist *p;
     /* set everything */
     set(uh, o);
 
+    if(iter) {
+      /* "part=item1 item2 item2" */
+      const char *part;
+      size_t plen;
+      const char *w;
+      size_t wlen;
+      char *sep;
+      bool urlencode = true;
+      const struct var *v;
+
+      if(!iinfo->ptr) {
+        part = iter->data;
+        sep = strchr(part, '=');
+        if(!sep)
+          errorf(ERROR_ITER, "wrong iterate syntax");
+        plen = sep - part;
+        if(sep[-1] == ':') {
+          urlencode = false;
+          plen--;
+        }
+        w = sep + 1;
+        /* store for next lap */
+        iinfo->part = part;
+        iinfo->plen = plen;
+        v = comp2var(part, plen);
+        if(!v)
+          errorf(ERROR_ITER, "bad component for iterate");
+        if(iinfo->varmask & (1<<v->part))
+          errorf(ERROR_ITER, "duplicate component for iterate: %s", v->name);
+      }
+      else {
+        part = iinfo->part;
+        plen = iinfo->plen;
+        v = comp2var(part, plen);
+        w = iinfo->ptr;
+      }
+
+      sep = strchr(w, ' ');
+      if(sep) {
+        wlen = sep - w;
+        iinfo->ptr = sep + 1; /* next word is here */
+      }
+      else {
+        /* last word */
+        wlen = strlen(w);
+        iinfo->ptr = NULL;
+      }
+      curl_msnprintf(iterbuf, sizeof(iterbuf), "%.*s%s=%.*s", (int)plen, part,
+                     urlencode ? "" : ":",
+                     (int)wlen, w);
+      setone(uh, o, iterbuf);
+      if(iter && iter->next) {
+        struct iterinfo info;
+        memset(&info, 0, sizeof(info));
+        info.uh = uh;
+        info.varmask = iinfo->varmask | (1 << v->part);
+        singleurl(o, url, &info, iter->next);
+      }
+    }
+
     /* append path segments */
     for(p = o->append_path; p; p = p->next) {
       char *apath = p->data;
@@ -856,7 +992,7 @@ static void singleurl(struct option *o,
       curl_url_get(uh, CURLUPART_PATH, &opath, 0);
 
       /* does the existing path end with a slash, then don't
-         add one inbetween */
+         add one in between */
       olen = strlen(opath);
 
       /* append the new segment */
@@ -886,7 +1022,9 @@ static void singleurl(struct option *o,
     /* put the query back */
     qpair2query(uh, o);
 
-    if(o->jsonout)
+    if(iter && iter->next)
+      ;
+    else if(o->jsonout)
       json(o, uh);
     else if(o->format) {
       /* custom output format */
@@ -909,6 +1047,9 @@ static void singleurl(struct option *o,
     freeqpairs();
 
     o->urls++;
+
+  } while(iinfo->ptr);
+  if(!iinfo->uh)
     curl_url_cleanup(uh);
 }
 
@@ -952,11 +1093,27 @@ int main(int argc, const char **argv)
         if(eol[-1] == '\r')
           /* CRLF detected */
           eol--;
-        *eol = 0; /* end of URL */
-        singleurl(&o, buffer);
       }
+      else if(feof(o.url))
+        /* end of file */
+        eol = strlen(buffer) + buffer;
       else {
-        /* no newline or no content, skip */
+        /* no newline but not end of file means that this line is truncated
+           and we are lost */
+        break;
+      }
+
+      /* trim trailing spaces and tabs */
+      while((eol > buffer) &&
+            ((eol[-1] == ' ') || eol[-1] == '\t'))
+        eol--;
+
+      if(eol > buffer) {
+        /* if there is actual content left to deal with */
+        struct iterinfo iinfo;
+        memset(&iinfo, 0, sizeof(iinfo));
+        *eol = 0; /* end of URL */
+        singleurl(&o, buffer, &iinfo, o.iter_list);
       }
     }
     if(o.urlopen)
@@ -968,11 +1125,16 @@ int main(int argc, const char **argv)
     do {
       if(node) {
         const char *url = node->data;
-        singleurl(&o, url);
+        struct iterinfo iinfo;
+        memset(&iinfo, 0, sizeof(iinfo));
+        singleurl(&o, url, &iinfo, o.iter_list);
         node = node->next;
       }
-      else
-        singleurl(&o, NULL);
+      else {
+        struct iterinfo iinfo;
+        memset(&iinfo, 0, sizeof(iinfo));
+        singleurl(&o, NULL, &iinfo, o.iter_list);
+      }
     } while(node);
   }
   if(o.jsonout)
@@ -980,6 +1142,7 @@ int main(int argc, const char **argv)
   /* we're done with libcurl, so clean it up */
   curl_slist_free_all(o.url_list);
   curl_slist_free_all(o.set_list);
+  curl_slist_free_all(o.iter_list);
   curl_slist_free_all(o.trim_list);
   curl_slist_free_all(o.append_path);
   curl_slist_free_all(o.append_query);
